@@ -4,6 +4,8 @@
 
 An asyncio-based library to communicate with SpamAssassin's SPAMD service."""
 
+from loguru import logger
+
 from .connections import Timeout
 from .frontend import (
     check,
@@ -19,7 +21,9 @@ from .options import ActionOption, MessageClassOption
 
 
 __author__ = "Michael Caley"
-__copyright__ = "Copyright 2016-2020 Michael Caley"
+__copyright__ = "Copyright 2016-2022 Michael Caley"
 __license__ = "MIT"
-__version__ = "0.7.1"
+__version__ = "0.8.0"
 __email__ = "mjcaley@darkarctic.com"
+
+logger.disable(__package__)
