@@ -1,122 +1,26 @@
-#!/usr/bin/end python
+#!/usr/bin/env python
 
 """Frontend functions for the package."""
 
-import logging
-import time
 from typing import (
     Any,
-    Callable,
     Dict,
-    NamedTuple,
     Optional,
     SupportsBytes,
-    Type,
     Union,
 )
-from ssl import SSLContext
 
-from .connections import Timeout, new_connection, new_ssl_context, ConnectionManager
-from .exceptions import BadResponse, ParseError
-from .header_values import CompressValue, MessageClassValue, UserValue
+from loguru import logger
+
+from .client import Client
+from .connections import Timeout
+from .header_values import MessageClassValue
 from .options import ActionOption, MessageClassOption
-from .incremental_parser import ResponseParser, parse_set_remove_value
+from .incremental_parser import parse_set_remove_value
 from .responses import Response
 from .requests import Request
 
 
-ConnectionFactory = Callable[
-    [
-        Optional[str],
-        Optional[int],
-        Optional[str],
-        Optional[Timeout],
-        Optional[SSLContext],
-    ],
-    ConnectionManager,
-]
-SSLFactory = Callable[[Any], Optional[SSLContext]]
-
-
-class Client(NamedTuple):
-    ssl_context_factory: SSLFactory
-    connection_factory: ConnectionFactory
-    parser_factory: Type[ResponseParser]
-
-
-DEFAULT_CLIENT = Client(new_ssl_context, new_connection, ResponseParser)
-
-LOGGER = logging.getLogger("aiospamc")
-
-
-async def request(
-    req: Request, connection: ConnectionManager, parser: ResponseParser
-) -> Response:
-    """Sends a request and returns the parsed response.
-
-    :param req: The request to send.
-    :param connection: An instance of a connection.
-    :param parser: An instance of a response parser.
-
-    :return: The parsed response.
-
-    :raises BadResponse: If the response from SPAMD is ill-formed this exception will be raised.
-    """
-
-    start = time.monotonic()
-    LOGGER.info(
-        "Sending %s request",
-        req.verb,
-        extra={
-            "message_id": id(req.body),
-            "request_id": id(req),
-            "connection_id": id(connection),
-        },
-    )
-    response = await connection.request(bytes(req))
-    try:
-        parsed_response = parser.parse(response)
-    except ParseError as error:
-        LOGGER.exception(
-            "Error parsing response",
-            exc_info=error,
-            extra={"request_id": id(req), "connection_id": id(connection)},
-        )
-        raise BadResponse(response) from error
-    response_obj = Response(**parsed_response)
-    response_obj.raise_for_status()
-    end = time.monotonic()
-    LOGGER.info(
-        "Successfully received response in %0.2f",
-        end - start,
-        extra={"request_id": id(req), "connection_id": id(connection)},
-    )
-
-    return response_obj
-
-
-def _new_connection(
-    connection_factory: ConnectionFactory,
-    ssl_context_factory: SSLFactory,
-    host: str = "localhost",
-    port: int = 783,
-    socket_path: str = None,
-    timeout: Timeout = None,
-    verify: Any = None,
-) -> ConnectionManager:
-    ssl_context = ssl_context_factory(verify)
-    connection = connection_factory(host, port, socket_path, timeout, ssl_context)
-
-    return connection
-
-
-def _add_headers(req: Request, user: Optional[str], compress: Optional[bool]) -> None:
-    if user:
-        req.headers["User"] = UserValue(user)
-    if compress:
-        req.headers["Compress"] = CompressValue()
-
-
 async def check(
     message: Union[bytes, SupportsBytes],
     *,
@@ -137,7 +41,7 @@ async def check(
     :param socket_path: Path to Unix socket.
     :param timeout: Timeout settings.
     :param verify:
-        Enable SSL. `True` will use the root certificates from the :module:`certifi` package.
+        Enable SSL. `True` will use the root certificates from the :py:mod:`certifi` package.
         `False` will use SSL, but not verify the root certificates. Passing a string to a filename
         will use the path to verify the root certificates.
     :param user: Username to pass to the SPAMD service.
@@ -168,22 +72,38 @@ async def check(
     :raises ClientTimeoutException: Client timed out during connection.
     """
 
-    client = kwargs.get("client", DEFAULT_CLIENT)
+    client = kwargs.get("client", Client())
 
     req = Request("CHECK", body=bytes(message))
-    _add_headers(req, user, compress)
-    connection = _new_connection(
-        client.connection_factory,
-        client.ssl_context_factory,
-        host,
-        port,
-        socket_path,
-        timeout,
-        verify,
+    context_logger = logger.bind(
+        host=host,
+        port=port,
+        socket_path=socket_path,
+        user=user,
+        request=req,
     )
-    parser = client.parser_factory()
+    context_logger.info("Sending CHECK request")
 
-    return await request(req, connection, parser)
+    try:
+        response = await client.request(
+            req,
+            host=host,
+            port=port,
+            socket_path=socket_path,
+            timeout=timeout,
+            verify=verify,
+            user=user,
+            compress=compress,
+        )
+    except Exception:
+        context_logger.exception("Exception when calling check function")
+        raise
+
+    context_logger.bind(response=response).success(
+        "Successfully completed check function"
+    )
+
+    return response
 
 
 async def headers(
@@ -206,7 +126,7 @@ async def headers(
     :param socket_path: Path to Unix socket.
     :param timeout: Timeout settings.
     :param verify:
-        Enable SSL. `True` will use the root certificates from the :module:`certifi` package.
+        Enable SSL. `True` will use the root certificates from the :py:mod:`certifi` package.
         `False` will use SSL, but not verify the root certificates. Passing a string to a filename
         will use the path to verify the root certificates.
     :param user: Username to pass to the SPAMD service.
@@ -238,22 +158,38 @@ async def headers(
     :raises ClientTimeoutException: Client timed out during connection.
     """
 
-    client = kwargs.get("client", DEFAULT_CLIENT)
+    client = kwargs.get("client", Client())
 
     req = Request("HEADERS", body=bytes(message))
-    _add_headers(req, user, compress)
-    connection = _new_connection(
-        client.connection_factory,
-        client.ssl_context_factory,
-        host,
-        port,
-        socket_path,
-        timeout,
-        verify,
+    context_logger = logger.bind(
+        host=host,
+        port=port,
+        socket_path=socket_path,
+        user=user,
+        request=req,
     )
-    parser = client.parser_factory()
+    context_logger.info("Sending HEADERS request")
 
-    return await request(req, connection, parser)
+    try:
+        response = await client.request(
+            req,
+            host=host,
+            port=port,
+            socket_path=socket_path,
+            timeout=timeout,
+            verify=verify,
+            user=user,
+            compress=compress,
+        )
+    except Exception:
+        context_logger.exception("Exception when calling headers function")
+        raise
+
+    context_logger.bind(response=response).success(
+        "Successfully completed headers function"
+    )
+
+    return response
 
 
 async def ping(
@@ -272,7 +208,7 @@ async def ping(
     :param socket_path: Path to Unix socket.
     :param timeout: Timeout settings.
     :param verify:
-        Enable SSL. `True` will use the root certificates from the :module:`certifi` package.
+        Enable SSL. `True` will use the root certificates from the :py:mod:`certifi` package.
         `False` will use SSL, but not verify the root certificates. Passing a string to a filename
         will use the path to verify the root certificates.
 
@@ -299,21 +235,35 @@ async def ping(
     :raises ClientTimeoutException: Client timed out during connection.
     """
 
-    client = kwargs.get("client", DEFAULT_CLIENT)
+    client = kwargs.get("client", Client())
 
     req = Request("PING")
-    connection = _new_connection(
-        client.connection_factory,
-        client.ssl_context_factory,
-        host,
-        port,
-        socket_path,
-        timeout,
-        verify,
+    context_logger = logger.bind(
+        host=host,
+        port=port,
+        socket_path=socket_path,
+        request=req,
     )
-    parser = client.parser_factory()
+    context_logger.info("Sending PING request")
 
-    return await request(req, connection, parser)
+    try:
+        response = await client.request(
+            req,
+            host=host,
+            port=port,
+            socket_path=socket_path,
+            timeout=timeout,
+            verify=verify,
+        )
+    except Exception:
+        context_logger.exception("Exception when calling ping function")
+        raise
+
+    context_logger.bind(response=response).success(
+        "Successfully completed ping function"
+    )
+
+    return response
 
 
 async def process(
@@ -336,7 +286,7 @@ async def process(
     :param socket_path: Path to Unix socket.
     :param timeout: Timeout settings.
     :param verify:
-        Enable SSL. `True` will use the root certificates from the :module:`certifi` package.
+        Enable SSL. `True` will use the root certificates from the :py:mod:`certifi` package.
         `False` will use SSL, but not verify the root certificates. Passing a string to a filename
         will use the path to verify the root certificates.
     :param user: Username to pass to the SPAMD service.
@@ -368,22 +318,38 @@ async def process(
     :raises ClientTimeoutException: Client timed out during connection.
     """
 
-    client = kwargs.get("client", DEFAULT_CLIENT)
+    client = kwargs.get("client", Client())
 
     req = Request("PROCESS", body=bytes(message))
-    _add_headers(req, user, compress)
-    connection = _new_connection(
-        client.connection_factory,
-        client.ssl_context_factory,
-        host,
-        port,
-        socket_path,
-        timeout,
-        verify,
+    context_logger = logger.bind(
+        host=host,
+        port=port,
+        socket_path=socket_path,
+        user=user,
+        request=req,
     )
-    parser = client.parser_factory()
+    context_logger.info("Sending PROCESS request")
 
-    return await request(req, connection, parser)
+    try:
+        response = await client.request(
+            req,
+            host=host,
+            port=port,
+            socket_path=socket_path,
+            timeout=timeout,
+            verify=verify,
+            user=user,
+            compress=compress,
+        )
+    except Exception:
+        context_logger.exception("Exception when calling process function")
+        raise
+
+    context_logger.bind(response=response).success(
+        "Successfully completed process function"
+    )
+
+    return response
 
 
 async def report(
@@ -406,7 +372,7 @@ async def report(
     :param socket_path: Path to Unix socket.
     :param timeout: Timeout settings.
     :param verify:
-        Enable SSL. `True` will use the root certificates from the :module:`certifi` package.
+        Enable SSL. `True` will use the root certificates from the :py:mod:`certifi` package.
         `False` will use SSL, but not verify the root certificates. Passing a string to a filename
         will use the path to verify the root certificates.
     :param user: Username to pass to the SPAMD service.
@@ -437,22 +403,38 @@ async def report(
     :raises ClientTimeoutException: Client timed out during connection.
     """
 
-    client = kwargs.get("client", DEFAULT_CLIENT)
+    client = kwargs.get("client", Client())
 
     req = Request("REPORT", body=bytes(message))
-    _add_headers(req, user, compress)
-    connection = _new_connection(
-        client.connection_factory,
-        client.ssl_context_factory,
-        host,
-        port,
-        socket_path,
-        timeout,
-        verify,
+    context_logger = logger.bind(
+        host=host,
+        port=port,
+        socket_path=socket_path,
+        user=user,
+        request=req,
     )
-    parser = client.parser_factory()
+    context_logger.info("Sending REPORT request")
 
-    return await request(req, connection, parser)
+    try:
+        response = await client.request(
+            req,
+            host=host,
+            port=port,
+            socket_path=socket_path,
+            timeout=timeout,
+            verify=verify,
+            user=user,
+            compress=compress,
+        )
+    except Exception:
+        context_logger.exception("Exception when calling report function")
+        raise
+
+    context_logger.bind(response=response).success(
+        "Successfully completed report function"
+    )
+
+    return response
 
 
 async def report_if_spam(
@@ -475,7 +457,7 @@ async def report_if_spam(
     :param socket_path: Path to Unix socket.
     :param timeout: Timeout settings.
     :param verify:
-        Enable SSL. `True` will use the root certificates from the :module:`certifi` package.
+        Enable SSL. `True` will use the root certificates from the :py:mod:`certifi` package.
         `False` will use SSL, but not verify the root certificates. Passing a string to a filename
         will use the path to verify the root certificates.
     :param user: Username to pass to the SPAMD service.
@@ -507,22 +489,38 @@ async def report_if_spam(
     :raises ClientTimeoutException: Client timed out during connection.
     """
 
-    client = kwargs.get("client", DEFAULT_CLIENT)
+    client = kwargs.get("client", Client())
 
     req = Request("REPORT_IFSPAM", body=bytes(message))
-    _add_headers(req, user, compress)
-    connection = _new_connection(
-        client.connection_factory,
-        client.ssl_context_factory,
-        host,
-        port,
-        socket_path,
-        timeout,
-        verify,
+    context_logger = logger.bind(
+        host=host,
+        port=port,
+        socket_path=socket_path,
+        user=user,
+        request=req,
     )
-    parser = client.parser_factory()
+    context_logger.info("Sending REPORT_IFSPAM request")
 
-    return await request(req, connection, parser)
+    try:
+        response = await client.request(
+            req,
+            host=host,
+            port=port,
+            socket_path=socket_path,
+            timeout=timeout,
+            verify=verify,
+            user=user,
+            compress=compress,
+        )
+    except Exception:
+        context_logger.exception("Exception when calling report_if_spam function")
+        raise
+
+    context_logger.bind(response=response).success(
+        "Successfully completed report_if_spam function"
+    )
+
+    return response
 
 
 async def symbols(
@@ -545,7 +543,7 @@ async def symbols(
     :param socket_path: Path to Unix socket.
     :param timeout: Timeout settings.
     :param verify:
-        Enable SSL. `True` will use the root certificates from the :module:`certifi` package.
+        Enable SSL. `True` will use the root certificates from the :py:mod:`certifi` package.
         `False` will use SSL, but not verify the root certificates. Passing a string to a filename
         will use the path to verify the root certificates.
     :param user: Username to pass to the SPAMD service.
@@ -577,22 +575,38 @@ async def symbols(
     :raises ClientTimeoutException: Client timed out during connection.
     """
 
-    client = kwargs.get("client", DEFAULT_CLIENT)
+    client = kwargs.get("client", Client())
 
     req = Request("SYMBOLS", body=bytes(message))
-    _add_headers(req, user, compress)
-    connection = _new_connection(
-        client.connection_factory,
-        client.ssl_context_factory,
-        host,
-        port,
-        socket_path,
-        timeout,
-        verify,
+    context_logger = logger.bind(
+        host=host,
+        port=port,
+        socket_path=socket_path,
+        user=user,
+        request=req,
     )
-    parser = client.parser_factory()
+    context_logger.info("Sending SYMBOLS request")
 
-    return await request(req, connection, parser)
+    try:
+        response = await client.request(
+            req,
+            host=host,
+            port=port,
+            socket_path=socket_path,
+            timeout=timeout,
+            verify=verify,
+            user=user,
+            compress=compress,
+        )
+    except Exception:
+        context_logger.exception("Exception when calling symbols function")
+        raise
+
+    context_logger.bind(response=response).success(
+        "Successfully completed symbols function"
+    )
+
+    return response
 
 
 async def tell(
@@ -621,7 +635,7 @@ async def tell(
     :param socket_path: Path to Unix socket.
     :param timeout: Timeout settings.
     :param verify:
-        Enable SSL. `True` will use the root certificates from the :module:`certifi` package.
+        Enable SSL. `True` will use the root certificates from the :py:mod:`certifi` package.
         `False` will use SSL, but not verify the root certificates. Passing a string to a filename
         will use the path to verify the root certificates.
     :param user: Username to pass to the SPAMD service.
@@ -652,7 +666,7 @@ async def tell(
     :raises ClientTimeoutException: Client timed out during connection.
     """
 
-    client = kwargs.get("client", DEFAULT_CLIENT)
+    client = kwargs.get("client", Client())
 
     headers: Dict[str, Any] = {
         "Message-class": MessageClassValue(MessageClassOption(message_class))
@@ -662,17 +676,32 @@ async def tell(
     if set_action:
         headers["Set"] = parse_set_remove_value(set_action)
     req = Request("TELL", headers=headers, body=bytes(message))
-    _add_headers(req, user, compress)
-
-    connection = _new_connection(
-        client.connection_factory,
-        client.ssl_context_factory,
-        host,
-        port,
-        socket_path,
-        timeout,
-        verify,
+    context_logger = logger.bind(
+        host=host,
+        port=port,
+        socket_path=socket_path,
+        user=user,
+        request=req,
     )
-    parser = client.parser_factory()
+    context_logger.info("Sending TELL request")
 
-    return await request(req, connection, parser)
+    try:
+        response = await client.request(
+            req,
+            host=host,
+            port=port,
+            socket_path=socket_path,
+            timeout=timeout,
+            verify=verify,
+            user=user,
+            compress=compress,
+        )
+    except Exception:
+        context_logger.exception("Exception when calling tell function")
+        raise
+
+    context_logger.bind(response=response).success(
+        "Successfully completed tell function"
+    )
+
+    return response
